@@ -5,7 +5,6 @@
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
-import time
 import math
 import os.path
 try:
@@ -13,7 +12,7 @@ try:
 except ImportError:
     from urllib.parse import unquote    #python3 @Reimport @UnresolvedImport
 
-from xpra.os_util import bytestostr, is_X11, WIN32, OSX, POSIX, PYTHON3
+from xpra.os_util import bytestostr, strtobytes, is_X11, monotonic_time, WIN32, OSX, POSIX, PYTHON3, PYTHON2
 from xpra.util import (
     AdHocStruct, typedict, envint, envbool, nonl, csv, first_time,
     WORKSPACE_UNSET, WORKSPACE_ALL, WORKSPACE_NAMES, MOVERESIZE_DIRECTION_STRING, SOURCE_INDICATION_STRING,
@@ -24,18 +23,20 @@ from xpra.util import (
     MOVERESIZE_SIZE_LEFT, MOVERESIZE_MOVE,
     )
 from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_cairo, is_gtk3
-from xpra.gtk_common.gobject_util import no_arg_signal
+from xpra.gtk_common.gobject_util import no_arg_signal, one_arg_signal
 from xpra.gtk_common.gtk_util import (
     get_xwindow, get_pixbuf_from_data, get_default_root_window,
     is_realized, display_get_default, drag_status,
     newTargetEntry, drag_context_targets, drag_context_actions,
     drag_dest_window, drag_widget_get_data,
+    enable_alpha,
     gio_File, query_info_async, load_contents_async, load_contents_finish,
     WINDOW_POPUP, WINDOW_TOPLEVEL, GRAB_STATUS_STRING, GRAB_SUCCESS,
     SCROLL_UP, SCROLL_DOWN, SCROLL_LEFT, SCROLL_RIGHT,
     DEST_DEFAULT_MOTION, DEST_DEFAULT_HIGHLIGHT, ACTION_COPY,
     BUTTON_PRESS_MASK, BUTTON_RELEASE_MASK, POINTER_MOTION_MASK,
     POINTER_MOTION_HINT_MASK, ENTER_NOTIFY_MASK, LEAVE_NOTIFY_MASK,
+    WINDOW_EVENT_MASK,
     )
 from xpra.gtk_common.keymap import KEY_TRANSLATIONS
 from xpra.client.client_window_base import ClientWindowBase
@@ -54,9 +55,9 @@ eventslog = Logger("events")
 shapelog = Logger("shape")
 mouselog = Logger("mouse")
 geomlog = Logger("geometry")
-menulog = Logger("menu")
 grablog = Logger("grab")
 draglog = Logger("dragndrop")
+alphalog = Logger("alpha")
 
 gtk     = import_gtk()
 gdk     = import_gdk()
@@ -65,11 +66,12 @@ cairo   = import_cairo()
 CAN_SET_WORKSPACE = False
 HAS_X11_BINDINGS = False
 USE_X11_BINDINGS = POSIX and envbool("XPRA_USE_X11_BINDINGS", is_X11())
-prop_get, prop_set = None, None
+prop_get, prop_set, prop_del = None, None, None
+NotifyInferior = None
 if USE_X11_BINDINGS:
     try:
-        from xpra.gtk_common.error import xsync, verify_sync
-        from xpra.x11.gtk_x11.prop import prop_get, prop_set
+        from xpra.gtk_common.error import xlog, verify_sync
+        from xpra.x11.gtk_x11.prop import prop_get, prop_set, prop_del
         from xpra.x11.bindings.window_bindings import constants, X11WindowBindings, SHAPE_KIND  #@UnresolvedImport
         from xpra.x11.bindings.core_bindings import X11CoreBindings, set_context_check
         from xpra.x11.gtk_x11.send_wm import send_wm_workspace
@@ -77,6 +79,7 @@ if USE_X11_BINDINGS:
         set_context_check(verify_sync)
         X11Window = X11WindowBindings()
         X11Core = X11CoreBindings()
+        NotifyInferior = constants["NotifyInferior"]
         HAS_X11_BINDINGS = True
 
         SubstructureNotifyMask = constants["SubstructureNotifyMask"]
@@ -122,7 +125,7 @@ def parse_padding_colors(colors_str):
     if colors_str:
         try:
             padding_colors = tuple(float(x.strip()) for x in colors_str.split(","))
-            assert len(PADDING_COLORS)==3, "you must specify 3 components"
+            assert len(padding_colors)==3, "you must specify 3 components"
         except Exception as e:
             log.warn("Warning: invalid padding colors specified,")
             log.warn(" %s", e)
@@ -190,6 +193,8 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
 
     __common_gsignals__ = {
         "state-updated"         : no_arg_signal,
+        "xpra-focus-out-event"  : one_arg_signal,
+        "xpra-focus-in-event"   : one_arg_signal,
         }
 
     #maximum size of the actual window:
@@ -202,13 +207,14 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         #gtk.Window.__init__() is called from do_init_window()
 
     def init_window(self, metadata):
-        self.init_max_window_size(metadata)
+        self.init_max_window_size()
         if self._is_popup(metadata):
             window_type = WINDOW_POPUP
         else:
             window_type = WINDOW_TOPLEVEL
         self.on_realize_cb = {}
         self.do_init_window(window_type)
+        self.init_drawing_area()
         self.set_decorated(self._is_decorated(metadata))
         self.set_app_paintable(True)
         self._window_state = {}
@@ -226,12 +232,38 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         #add platform hooks
         self.connect_after("realize", self.on_realize)
         self.connect('unrealize', self.on_unrealize)
-        self.add_events(self.WINDOW_EVENT_MASK)
-        if DRAGNDROP:
+        self.add_events(WINDOW_EVENT_MASK)
+        if DRAGNDROP and not self._client.readonly:
             self.init_dragndrop()
         self.init_focus()
         ClientWindowBase.init_window(self, metadata)
 
+    def init_drawing_area(self):
+        widget = gtk.DrawingArea()
+        widget.show()
+        self.drawing_area = widget
+        self.init_widget_events(widget)
+        self.add(widget)
+
+    def init_widget_events(self, widget):
+        widget.add_events(WINDOW_EVENT_MASK)
+        def motion(_w, event):
+            self._do_motion_notify_event(event)
+            return True
+        widget.connect("motion-notify-event", motion)
+        def press(_w, event):
+            self._do_button_press_event(event)
+            return True
+        widget.connect("button-press-event", press)
+        def release(_w, event):
+            self._do_button_release_event(event)
+            return True
+        widget.connect("button-release-event", release)
+        def scroll(_w, event):
+            self._do_scroll_event(event)
+            return True
+        widget.connect("scroll-event", scroll)
+
 
     ######################################################################
     # drag and drop:
@@ -278,8 +310,10 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         dest_window = xid(drag_dest_window(context))
         source_window = xid(context.get_source_window())
         suggested_action = context.get_suggested_action()
-        draglog("drag_got_data_cb context: source_window=%#x, dest_window=%#x, suggested_action=%s, actions=%s, targets=%s",
-                source_window, dest_window, suggested_action, actions, targets)
+        draglog("drag_got_data_cb context: source_window=%#x, dest_window=%#x",
+                source_window, dest_window)
+        draglog("drag_got_data_cb context: suggested_action=%s, actions=%s, targets=%s",
+                suggested_action, actions, targets)
         dtype = selection.get_data_type()
         fmt = selection.get_format()
         l = selection.get_length()
@@ -413,6 +447,11 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
 
     def do_xpra_focus_out_event(self, event):
         focuslog("do_xpra_focus_out_event(%s)", event)
+        if NotifyInferior is not None:
+            detail = getattr(event, "detail", None)
+            if detail==NotifyInferior:
+                log("dropped NotifyInferior focus event")
+                return True
         self._focus_latest = False
         return self.schedule_recheck_focus()
 
@@ -422,7 +461,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         return self.schedule_recheck_focus()
 
 
-    def init_max_window_size(self, metadata):
+    def init_max_window_size(self):
         """ used by GL windows to enforce a hard limit on window sizes """
         saved_mws = self.max_window_size
         def clamp_to(maxw, maxh):
@@ -442,8 +481,10 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         #we have to take scaling into account (if any):
         clamp_to(*self._client.sp(*self.MAX_BACKING_DIMS))
         if self.max_window_size!=saved_mws:
-            log("init_max_window_size(..) max-window-size changed from %s to %s, because of max viewport dims %s and max backing dims %s",
-                saved_mws, self.max_window_size, self.MAX_VIEWPORT_DIMS, self.MAX_BACKING_DIMS)
+            log("init_max_window_size(..) max-window-size changed from %s to %s",
+                saved_mws, self.max_window_size)
+            log(" because of max viewport dims %s and max backing dims %s",
+                self.MAX_VIEWPORT_DIMS, self.MAX_BACKING_DIMS)
 
 
     def is_awt(self, metadata):
@@ -521,6 +562,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
 
 
     def setup_window(self, *args):
+        log("setup_window%s", args)
         self.set_alpha()
 
         if self._override_redirect:
@@ -718,7 +760,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         #try to enable alpha on this window if needed,
         #and if the backing class can support it:
         bc = self.get_backing_class()
-        metalog("set_alpha() has_alpha=%s, %s.HAS_ALPHA=%s, realized=%s",
+        alphalog("set_alpha() has_alpha=%s, %s.HAS_ALPHA=%s, realized=%s",
                 self._has_alpha, bc, bc.HAS_ALPHA, self.is_realized())
         #by default, only RGB (no transparency):
         #rgb_formats = tuple(BACKING_CLASS.RGB_MODES)
@@ -727,10 +769,11 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self._client_properties["encoding.transparency"] = False
             return
         if self._has_alpha and not self.is_realized():
-            if self.enable_alpha():
+            if enable_alpha(self):
                 self._client_properties["encodings.rgb_formats"] = ["RGBA", "RGB", "RGBX"]
                 self._window_alpha = True
             else:
+                alphalog("enable_alpha()=False")
                 self._has_alpha = False
                 self._client_properties["encoding.transparency"] = False
 
@@ -774,18 +817,24 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             #or when we get the configure event - which should come straight after
             #if we're changing the maximized state
             state_updates["maximized"] = bool(event.new_window_state & self.WINDOW_STATE_MAXIMIZED)
+        if event.changed_mask & self.WINDOW_STATE_FOCUSED:
+            state_updates["focused"] = bool(event.new_window_state & self.WINDOW_STATE_FOCUSED)
         self.update_window_state(state_updates)
 
     def update_window_state(self, state_updates):
+        if self._client.readonly:
+            log("update_window_state(%s) ignored in readonly mode", state_updates)
+            return
         if state_updates.get("maximized") is False or state_updates.get("fullscreen") is False:
             #if we unfullscreen or unmaximize, re-calculate offsets if we have any:
             w, h = self._backing.render_size
             ww, wh = self.get_size()
-            log("update_window_state(%s) unmax / unfullscreen - window_offset=%s, backing render_size=%s, window size=%s",
-                state_updates, self.window_offset, (w, h), (ww, wh))
+            log("update_window_state(%s) unmax or unfullscreen", state_updates)
+            log("window_offset=%s, backing render_size=%s, window size=%s",
+                self.window_offset, (w, h), (ww, wh))
             if self._backing.offsets!=(0, 0, 0, 0):
                 self.center_backing(w, h)
-                self.queue_draw(0, 0, ww, wh)
+                self.queue_draw_area(0, 0, ww, wh)
         #decide if this is really an update by comparing with our local state vars:
         #(could just be a notification of a state change we already know about)
         actual_updates = {}
@@ -837,6 +886,8 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
 
     def schedule_send_iconify(self):
         #calculate a good delay to prevent races causing minimize/unminimize loops:
+        if self._client.readonly:
+            return
         delay = 150
         spl = tuple(self._client.server_ping_latency)
         if spl:
@@ -876,6 +927,20 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         self.when_realized("command", do_set_command)
 
 
+    def set_x11_property(self, prop_name, dtype, dformat, value):
+        metalog("set_x11_property%s", (prop_name, dtype, dformat, value))
+        gdk_window = self.get_window()
+        if not dtype and not dformat:
+            #remove prop
+            prop_del(gdk_window, prop_name)
+            return
+        dtype = bytestostr(dtype)
+        if dtype=="latin1":
+            value = bytestostr(value)
+            if PYTHON2:
+               value = unicode(value)
+        prop_set(gdk_window, prop_name, dtype, value)
+
     def set_class_instance(self, wmclass_name, wmclass_class):
         if not self.is_realized():
             #Warning: window managers may ignore the icons we try to set
@@ -884,8 +949,8 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self.set_wmclass(wmclass_name, wmclass_class)
         elif HAS_X11_BINDINGS:
             xid = get_xwindow(self.get_window())
-            with xsync:
-                X11Window.setClassHint(xid, wmclass_class, wmclass_name)
+            with xlog:
+                X11Window.setClassHint(xid, strtobytes(wmclass_class), strtobytes(wmclass_name))
                 log("XSetClassHint(%s, %s) done", wmclass_class, wmclass_name)
 
     def set_shape(self, shape):
@@ -905,7 +970,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
                     #too expensive to log with actual rectangles:
                     shapelog("XShapeCombineRectangles(%#x, %s, %i, %i, %i rects)",
                              xid, name, x_off, y_off, len(rectangles))
-                    with xsync:
+                    with xlog:
                         X11Window.XShapeCombineRectangles(xid, kind, x_off, y_off, rectangles)
         self.when_realized("shape", do_set_shape)
 
@@ -1066,61 +1131,66 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
 
 
     def property_changed(self, widget, event):
-        statelog("property_changed(%s, %s) : %s", widget, event, event.atom)
-        if event.atom=="_NET_WM_DESKTOP" and self._been_mapped and not self._override_redirect and self._can_set_workspace:
-            self.do_workspace_changed(event)
-        elif event.atom=="_NET_FRAME_EXTENTS" and prop_get:
-            v = prop_get(self.get_window(), "_NET_FRAME_EXTENTS", ["u32"], ignore_errors=False)
-            statelog("_NET_FRAME_EXTENTS: %s", v)
-            if v:
-                if v==self._current_frame_extents:
-                    #unchanged
-                    return
-                if not self._been_mapped:
-                    #map event will take care of sending it
-                    return
-                if self.is_OR() or self.is_tray():
-                    #we can't do it: the server can't handle configure packets for OR windows!
-                    return
-                if not self._client.server_window_frame_extents:
-                    #can't send cheap "skip-geometry" packets or frame-extents feature not supported:
-                    return
-                #tell server about new value:
-                self._current_frame_extents = v
-                statelog("sending configure event to update _NET_FRAME_EXTENTS to %s", v)
-                self._window_state["frame"] = self._client.crect(*v)
-                self.send_configure_event(True)
-        elif event.atom=="XKLAVIER_STATE" and prop_get:
-            #unused for now, but log it:
-            xklavier_state = prop_get(self.get_window(), "XKLAVIER_STATE", ["integer"], ignore_errors=False)
-            keylog("XKLAVIER_STATE=%s", [hex(x) for x in (xklavier_state or [])])
-        elif event.atom=="_NET_WM_STATE" and prop_get:
-            wm_state_atoms = prop_get(self.get_window(), "_NET_WM_STATE", ["atom"], ignore_errors=False)
-            #code mostly duplicated from gtk_x11/window.py:
-            WM_STATE_NAME = {
-                             "fullscreen"            : ("_NET_WM_STATE_FULLSCREEN", ),
-                             "maximized"             : ("_NET_WM_STATE_MAXIMIZED_VERT", "_NET_WM_STATE_MAXIMIZED_HORZ"),
-                             "shaded"                : ("_NET_WM_STATE_SHADED", ),
-                             "sticky"                : ("_NET_WM_STATE_STICKY", ),
-                             "skip-pager"            : ("_NET_WM_STATE_SKIP_PAGER", ),
-                             "skip-taskbar"          : ("_NET_WM_STATE_SKIP_TASKBAR", ),
-                             "above"                 : ("_NET_WM_STATE_ABOVE", ),
-                             "below"                 : ("_NET_WM_STATE_BELOW", ),
-                             "focused"               : ("_NET_WM_STATE_FOCUSED", ),
-                             }
-            state_atoms = set(wm_state_atoms or [])
-            state_updates = {}
-            for state, atoms in WM_STATE_NAME.items():
-                var = "_" + state.replace("-", "_")           #ie: "skip-pager" -> "_skip_pager"
-                cur_state = getattr(self, var)
-                wm_state_is_set = set(atoms).issubset(state_atoms)
-                if wm_state_is_set and not cur_state:
-                    state_updates[state] = True
-                elif cur_state and not wm_state_is_set:
-                    state_updates[state] = False
-            log("_NET_WM_STATE=%s, state_updates=%s", wm_state_atoms, state_updates)
-            if state_updates:
-                self.update_window_state(state_updates)
+        atom = str(event.atom)
+        statelog("property_changed(%s, %s) : %s", widget, event, atom)
+        if atom=="_NET_WM_DESKTOP":
+            if self._been_mapped and not self._override_redirect and self._can_set_workspace:
+                self.do_workspace_changed(event)
+        elif atom=="_NET_FRAME_EXTENTS":
+            if prop_get:
+                v = prop_get(self.get_window(), "_NET_FRAME_EXTENTS", ["u32"], ignore_errors=False)
+                statelog("_NET_FRAME_EXTENTS: %s", v)
+                if v:
+                    if v==self._current_frame_extents:
+                        #unchanged
+                        return
+                    if not self._been_mapped:
+                        #map event will take care of sending it
+                        return
+                    if self.is_OR() or self.is_tray():
+                        #we can't do it: the server can't handle configure packets for OR windows!
+                        return
+                    if not self._client.server_window_frame_extents:
+                        #can't send cheap "skip-geometry" packets or frame-extents feature not supported:
+                        return
+                    #tell server about new value:
+                    self._current_frame_extents = v
+                    statelog("sending configure event to update _NET_FRAME_EXTENTS to %s", v)
+                    self._window_state["frame"] = self._client.crect(*v)
+                    self.send_configure_event(True)
+        elif atom=="XKLAVIER_STATE":
+            if prop_get:
+                #unused for now, but log it:
+                xklavier_state = prop_get(self.get_window(), "XKLAVIER_STATE", ["integer"], ignore_errors=False)
+                keylog("XKLAVIER_STATE=%s", [hex(x) for x in (xklavier_state or [])])
+        elif atom=="_NET_WM_STATE":
+            if prop_get:
+                wm_state_atoms = prop_get(self.get_window(), "_NET_WM_STATE", ["atom"], ignore_errors=False)
+                #code mostly duplicated from gtk_x11/window.py:
+                WM_STATE_NAME = {
+                    "fullscreen"    : ("_NET_WM_STATE_FULLSCREEN", ),
+                    "maximized"     : ("_NET_WM_STATE_MAXIMIZED_VERT", "_NET_WM_STATE_MAXIMIZED_HORZ"),
+                    "shaded"        : ("_NET_WM_STATE_SHADED", ),
+                    "sticky"        : ("_NET_WM_STATE_STICKY", ),
+                    "skip-pager"    : ("_NET_WM_STATE_SKIP_PAGER", ),
+                    "skip-taskbar"  : ("_NET_WM_STATE_SKIP_TASKBAR", ),
+                    "above"         : ("_NET_WM_STATE_ABOVE", ),
+                    "below"         : ("_NET_WM_STATE_BELOW", ),
+                    "focused"       : ("_NET_WM_STATE_FOCUSED", ),
+                    }
+                state_atoms = set(wm_state_atoms or [])
+                state_updates = {}
+                for state, atoms in WM_STATE_NAME.items():
+                    var = "_" + state.replace("-", "_")           #ie: "skip-pager" -> "_skip_pager"
+                    cur_state = getattr(self, var)
+                    wm_state_is_set = set(atoms).issubset(state_atoms)
+                    if wm_state_is_set and not cur_state:
+                        state_updates[state] = True
+                    elif cur_state and not wm_state_is_set:
+                        state_updates[state] = False
+                log("_NET_WM_STATE=%s, state_updates=%s", wm_state_atoms, state_updates)
+                if state_updates:
+                    self.update_window_state(state_updates)
 
 
     ######################################################################
@@ -1155,12 +1225,13 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             workspacelog("window is on all workspaces")
             suspend_resume = False
         elif desktop_workspace!=window_workspace:
-            workspacelog("window is on a different workspace, increasing its batch delay (desktop: %s, window: %s)",
-                         wn(desktop_workspace), wn(window_workspace))
+            workspacelog("window is on a different workspace, increasing its batch delay")
+            workspacelog(" desktop: %s, window: %s", wn(desktop_workspace), wn(window_workspace))
             suspend_resume = True
         elif self._window_workspace!=self._desktop_workspace:
             assert desktop_workspace==window_workspace
-            workspacelog("window was on a different workspace, resetting its batch delay (was desktop: %s, window: %s, now both on %s)",
+            workspacelog("window was on a different workspace, resetting its batch delay")
+            workspacelog(" (was desktop: %s, window: %s, now both on %s)",
                          wn(self._window_workspace), wn(self._desktop_workspace), wn(desktop_workspace))
             suspend_resume = False
         self._window_workspace = window_workspace
@@ -1195,8 +1266,9 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         desktop = self.get_desktop_workspace()
         ndesktops = self.get_workspace_count()
         current = self.get_window_workspace()
-        workspacelog("set_workspace(%s) realized=%s, current workspace=%s, detected=%s, desktop workspace=%s, ndesktops=%s",
-                     wn(workspace), self.is_realized(), wn(self._window_workspace), wn(current), wn(desktop), ndesktops)
+        workspacelog("set_workspace(%s) realized=%s", wn(workspace), self.is_realized())
+        workspacelog(" current workspace=%s, detected=%s, desktop workspace=%s, ndesktops=%s",
+                     wn(self._window_workspace), wn(current), wn(desktop), ndesktops)
         if not self._can_set_workspace or ndesktops is None:
             return
         if workspace==desktop or workspace==WORKSPACE_ALL or desktop is None:
@@ -1221,7 +1293,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         workspacelog("do_set_workspace: gdkwindow: %#x, mapped=%s, visible=%s",
                      get_xwindow(gdkwin), self.is_mapped(), gdkwin.is_visible())
         root = get_default_root_window()
-        with xsync:
+        with xlog:
             send_wm_workspace(root, gdkwin, workspace)
 
     def get_desktop_workspace(self):
@@ -1291,7 +1363,8 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
                 args, self.get_window(), GRAB_STATUS_STRING.get(r), self._client.pointer_grabbed)
 
     def pointer_ungrab(self, *args):
-        grablog("pointer_ungrab%s", args)
+        grablog("pointer_ungrab%s pointer_grabbed=%s",
+                args, self._client.pointer_grabbed)
         self._client.pointer_grabbed = False
         gdkwin = self.get_window()
         if gdkwin:
@@ -1301,8 +1374,9 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         return True
 
     def toggle_pointer_grab(self):
-        grablog("toggle_pointer_grab()")
-        if self._client.pointer_grabbed:
+        pg = self._client.pointer_grabbed
+        grablog("toggle_pointer_grab() pointer_grabbed=%s", pg)
+        if pg:
             self.pointer_ungrab()
         else:
             self.pointer_grab()
@@ -1316,24 +1390,6 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self.fullscreen()
 
 
-    def set_menu(self, menu):
-        menulog("set_menu(%s)", menu)
-        def do_set_menu():
-            self._client.set_window_menu(True, self._id, menu, self.application_action_callback, self.window_action_callback)
-        self.when_realized("menu", do_set_menu)
-
-    def application_action_callback(self, action_service, action, state, pdata):
-        self.call_action("application", action, state, pdata)
-
-    def window_action_callback(self, action_service, action, state, pdata):
-        self.call_action("window", action, state, pdata)
-
-    def call_action(self, action_type, action, state, pdata):
-        menulog("call_action%s", (action_type, action, state, pdata))
-        rpc_args = [action_type, self._id, action, state, pdata]
-        self._client.rpc_call("menu", rpc_args)
-
-
     ######################################################################
     # pointer overlay handling
     def cancel_remove_pointer_overlay_timer(self):
@@ -1394,7 +1450,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         if value:
             #repaint the scale value (in window coordinates):
             x, y, w, h = abs_coords(*value[2:5])
-            self.queue_draw(x, y, w, h)
+            self.queue_draw_area(x, y, w, h)
             #clear it shortly after:
             self.cancel_remove_pointer_overlay_timer()
             def remove_pointer_overlay():
@@ -1403,16 +1459,25 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self.remove_pointer_overlay_timer = self.timeout_add(CURSOR_IDLE_TIMEOUT*1000, remove_pointer_overlay)
         if prev:
             x, y, w, h = abs_coords(*prev[2:5])
-            self.queue_draw(x, y, w, h)
+            self.queue_draw_area(x, y, w, h)
 
 
+    def _do_button_press_event(self, event):
+        #gtk.Window.do_button_press_event(self, event)
+        self._button_action(event.button, event, True)
+
+    def _do_button_release_event(self, event):
+        #gtk.Window.do_button_release_event(self, event)
+        self._button_action(event.button, event, False)
+
     ######################################################################
     # pointer motion
 
-    def do_motion_notify_event(self, event):
+    def _do_motion_notify_event(self, event):
+        #gtk.Window.do_motion_notify_event(self, event)
         if self.moveresize_event:
             self.motion_moveresize(event)
-        ClientWindowBase.do_motion_notify_event(self, event)
+        ClientWindowBase._do_motion_notify_event(self, event)
 
     def motion_moveresize(self, event):
         x_root, y_root, direction, button, start_buttons, wx, wy, ww, wh = self.moveresize_event
@@ -1497,6 +1562,12 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             x, y = int(move[0]), int(move[1])
         if resize:
             w, h = int(resize[0]), int(resize[1])
+            if self._client.readonly:
+                #change size-constraints first,
+                #so the resize can be honoured:
+                sc = self._force_size_constraint(w, h)
+                self._metadata.update(sc)
+                self.set_metadata(sc)
         if move and resize:
             self.get_window().move_resize(x, y, w, h)
         elif move:
@@ -1529,7 +1600,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         root = self.get_window().get_screen().get_root_window()
         root_xid = get_xwindow(root)
         xwin = get_xwindow(self.get_window())
-        with xsync:
+        with xlog:
             X11Core.UngrabPointer()
             X11Window.sendClientMessage(root_xid, xwin, False, event_mask, "_NET_WM_MOVERESIZE",
                   x_root, y_root, direction, button, source_indication)
@@ -1613,7 +1684,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         context.set_line_cap(cairo.LINE_CAP_ROUND)
         context.translate(w/2, h/2)
         from xpra.client.spinner import cv
-        count = int(time.time()*4.0)
+        count = int(monotonic_time()*4.0)
         for i in range(8):      #8 lines
             context.set_source_rgba(0, 0, 0, cv.trs[count%8][i])
             context.move_to(0.0, -dim/4.0)
@@ -1628,7 +1699,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         #with normal windows, we just queue a draw request
         #and let the expose event paint the spinner
         w, h = self.get_size()
-        self.queue_draw(0, 0, w, h)
+        self.queue_draw_area(0, 0, w, h)
 
 
     def do_map_event(self, event):
@@ -1641,7 +1712,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self.process_map_event()
 
     def process_map_event(self):
-        x, y, w, h = self.get_window_geometry()
+        x, y, w, h = self.get_drawing_area_geometry()
         state = self._window_state
         props = self._client_properties
         self._client_properties = {}
@@ -1711,7 +1782,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
 
     def process_configure_event(self, skip_geometry=False):
         assert skip_geometry or not self.is_OR()
-        x, y, w, h = self.get_window_geometry()
+        x, y, w, h = self.get_drawing_area_geometry()
         w = max(1, w)
         h = max(1, h)
         ox, oy = self._pos
@@ -1730,11 +1801,11 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self._set_backing_size(w, h)
         elif self._backing and not self._iconified:
             geomlog("configure event: size unchanged, queueing redraw")
-            self.queue_draw(0, 0, w, h)
+            self.queue_draw_area(0, 0, w, h)
 
     def send_configure_event(self, skip_geometry=False):
         assert skip_geometry or not self.is_OR()
-        x, y, w, h = self.get_window_geometry()
+        x, y, w, h = self.get_drawing_area_geometry()
         w = max(1, w)
         h = max(1, h)
         state = self._window_state
@@ -1781,7 +1852,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         self._resize_counter = resize_counter
         if (w, h)==(ww, wh):
             self._backing.offsets = 0, 0, 0, 0
-            self.queue_draw(0, 0, w, h)
+            self.queue_draw_area(0, 0, w, h)
             return
         if not self._fullscreen and not self._maximized:
             gtk.Window.resize(self, w, h)
@@ -1791,7 +1862,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             self.center_backing(w, h)
         geomlog("backing offsets=%s, window offset=%s", self._backing.offsets, self.window_offset)
         self._set_backing_size(w, h)
-        self.queue_draw(0, 0, ww, wh)
+        self.queue_draw_area(0, 0, ww, wh)
 
     def center_backing(self, w, h):
         ww, wh = self.get_size()
@@ -1846,8 +1917,8 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             y += self.window_offset[1]
             #TODO: check this doesn't move it off-screen!
         self._resize_counter = resize_counter
-        window = self.get_window()
-        if window.get_position()==(x, y):
+        wx, wy = self.get_drawing_area_geometry()[:2]
+        if (wx, wy)==(x, y):
             #same location, just resize:
             if self._size==(w, h):
                 geomlog("window unchanged")
@@ -1861,6 +1932,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
             geomlog("window was not realized yet")
             self.realize()
         #adjust for window frame:
+        window = self.get_window()
         ox, oy = window.get_origin()[-2:]
         rx, ry = window.get_root_origin()
         ax = x - (ox - rx)
@@ -1895,14 +1967,12 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
     def noop_destroy(self):
         log.warn("Warning: window destroy called twice!")
 
-    def destroy(self):
+    def destroy(self):      #pylint: disable=method-hidden
         self.cancel_window_state_timer()
         self.cancel_send_iconifiy_timer()
         self.cancel_show_pointer_overlay_timer()
         self.cancel_remove_pointer_overlay_timer()
         self.cancel_focus_timer()
-        if self._client._set_window_menu:
-            self._client.set_window_menu(False, self._id, {})
         mrt = self.moveresize_timer
         if mrt:
             self.moveresize_timer = None
@@ -1919,8 +1989,14 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         self._unfocus()
         if not self._override_redirect:
             self.send("unmap-window", self._id, False)
+        if not PYTHON3:
+            try:
+                gtk.Window.do_unmap_event(self, event)
+            except Exception:
+                log.error("Error handling unmap event", exc_info=True)
 
     def do_delete_event(self, event):
+        #gtk.Window.do_delete_event(self, event)
         eventslog("do_delete_event(%s)", event)
         self._client.window_close_event(self._id)
         return True
@@ -1943,7 +2019,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         rx, ry = self._get_relative_pointer(event)
         #adjust for window offset:
         pointer = self._offset_pointer(x, y)
-        relative_pointer = self._offset_pointer(rx, ry)
+        relative_pointer = self._client.cp(rx, ry)
         #FIXME: state is used for both mods and buttons??
         modifiers = self._client.mask_to_names(event.state)
         buttons = self._event_buttons(event)
@@ -2013,7 +2089,7 @@ class GTKClientWindowBase(ClientWindowBase, gtk.Window):
         return ClientWindowBase.get_mouse_event_wid(self, x, y)
 
 
-    def do_scroll_event(self, event):
+    def _do_scroll_event(self, event):
         if self._client.readonly:
             return
         button_mapping = GDK_SCROLL_MAP.get(event.direction, -1)
