@@ -14,17 +14,20 @@ from xpra.x11.bindings.core_bindings import set_context_check, X11CoreBindings
 from xpra.x11.bindings.randr_bindings import RandRBindings  #@UnresolvedImport
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
 from xpra.x11.bindings.window_bindings import X11WindowBindings #@UnresolvedImport
-from xpra.gtk_common.error import XError, xswallow, xsync, trap, verify_sync
+from xpra.gtk_common.error import XError, xswallow, xsync, xlog, trap, verify_sync
 from xpra.gtk_common.gtk_util import get_xwindow, display_get_default, get_default_root_window
 from xpra.server.server_uuid import save_uuid, get_uuid
 from xpra.x11.fakeXinerama import find_libfakeXinerama, save_fakeXinerama_config, cleanup_fakeXinerama
 from xpra.x11.gtk_x11.prop import prop_get, prop_set
+from xpra.x11.gtk_x11.gdk_display_source import close_gdk_display_source
+from xpra.x11.gtk_x11.gdk_bindings import init_x11_filter, cleanup_x11_filter, cleanup_all_event_receivers
 from xpra.x11.common import MAX_WINDOW_SIZE
-from xpra.os_util import StringIOClass, monotonic_time, strtobytes, bytestostr, PYTHON3
+from xpra.os_util import monotonic_time, strtobytes, bytestostr, PYTHON3
 from xpra.util import engs, csv, typedict, iround, envbool, XPRA_DPI_NOTIFICATION_ID
 from xpra.net.compression import Compressed
 from xpra.server.gtk_server_base import GTKServerBase
 from xpra.x11.xkbhelper import clean_keyboard_state
+from xpra.scripts.config import FALSE_OPTIONS
 from xpra.log import Logger
 
 gdk = import_gdk()
@@ -89,25 +92,28 @@ class X11ServerCore(GTKServerBase):
         self.pointer_device_map = {}
         self.keys_pressed = {}
         self.last_mouse_user = None
+        self.x11_filter = False
         GTKServerBase.__init__(self)
+        log("XShape=%s", X11Window.displayHasXShape())
 
     def init(self, opts):
         self.do_init(opts)
         GTKServerBase.init(self, opts)
-        self.features_init()
 
-    def features_init(self):
+    def server_init(self):
         with xsync:
             self.x11_init()
         from xpra.server import server_features
         if server_features.windows:
             from xpra.x11.x11_window_filters import init_x11_window_filters
             init_x11_window_filters()
+        super(X11ServerCore, self).server_init()
 
     def do_init(self, opts):
+        self.opengl = opts.opengl
         self.randr = opts.resize_display
         self.randr_exact_size = False
-        self.fake_xinerama = False      #only enabled in seamless server
+        self.fake_xinerama = "no"      #only enabled in seamless server
         self.current_xinerama_config = None
         #x11 keyboard bits:
         self.current_keyboard_group = None
@@ -115,10 +121,12 @@ class X11ServerCore(GTKServerBase):
 
     def x11_init(self):
         clean_keyboard_state()
-        if self.fake_xinerama:
-            self.libfakeXinerama_so = find_libfakeXinerama()
-        else:
+        if self.fake_xinerama in FALSE_OPTIONS:
             self.libfakeXinerama_so = None
+        elif os.path.isabs(self.fake_xinerama):
+            self.libfakeXinerama_so = self.fake_xinerama
+        else:
+            self.libfakeXinerama_so = find_libfakeXinerama()
         if not X11Keyboard.hasXFixes() and self.cursors:
             log.error("Error: cursor forwarding support disabled")
         if not X11Keyboard.hasXTest():
@@ -130,6 +138,9 @@ class X11ServerCore(GTKServerBase):
             self.init_randr()
         self.init_cursor()
         self.query_opengl()
+        self.x11_filter = init_x11_filter()
+        assert self.x11_filter
+
 
     def init_randr(self):
         self.randr = RandR.has_randr()
@@ -152,19 +163,10 @@ class X11ServerCore(GTKServerBase):
                 self.randr = False
                 self.randr_exact_size = False
         screenlog("randr=%s, exact size=%s", self.randr, self.randr_exact_size)
-        display = display_get_default()
-        i=0
-        while i<display.get_n_screens():
-            screen = display.get_screen(i)
-            screen.connect("size-changed", self._screen_size_changed)
-            i += 1
         screenlog("randr enabled: %s", self.randr)
         if not self.randr:
             screenlog.warn("Warning: no X11 RandR support on %s", os.environ.get("DISPLAY"))
 
-    def _screen_size_changed(self, screen):
-        pass
-
 
     def init_cursor(self):
         #cursor:
@@ -179,12 +181,15 @@ class X11ServerCore(GTKServerBase):
         X11Keyboard.selectCursorChange(True)
 
     def get_display_bit_depth(self):
-        with xswallow:
+        with xlog:
             return X11Window.get_depth(X11Window.getDefaultRootWindow())
         return 0
 
     def query_opengl(self):
         self.opengl_props = {}
+        if self.opengl.lower()=="noprobe":
+            gllog("query_opengl() skipped: %s", self.opengl)
+            return
         blacklisted_kernel_modules = []
         for mod in ("vboxguest", "vboxvideo"):
             if os.path.exists("/sys/module/%s" % mod):
@@ -215,7 +220,7 @@ class X11ServerCore(GTKServerBase):
                         v = bytestostr(parts[1].strip())
                         self.opengl_props[k] = v
                     gllog("opengl props=%s", self.opengl_props)
-                    gllog.info(" OpenGL is supported on display '%s'", os.environ.get("DISPLAY"))
+                    gllog.info("OpenGL is supported on display '%s'", os.environ.get("DISPLAY"))
                     renderer = self.opengl_props.get("renderer")
                     if renderer:
                         gllog.info(" using '%s' renderer", renderer)
@@ -285,8 +290,8 @@ class X11ServerCore(GTKServerBase):
 
     def init_packet_handlers(self):
         GTKServerBase.init_packet_handlers(self)
-        self._authenticated_ui_packet_handlers["force-ungrab"] = self._process_force_ungrab
-        self._authenticated_ui_packet_handlers["wheel-motion"] = self._process_wheel_motion
+        self.add_packet_handler("force-ungrab", self._process_force_ungrab)
+        self.add_packet_handler("wheel-motion", self._process_wheel_motion)
 
 
     def init_virtual_devices(self, _devices):
@@ -301,11 +306,28 @@ class X11ServerCore(GTKServerBase):
         return env
 
     def do_cleanup(self):
-        GTKServerBase.do_cleanup(self)
+        if self.x11_filter:
+            self.x11_filter = False
+            cleanup_x11_filter()
+            #try a few times:
+            #errors happen because windows are being destroyed
+            #(even more so when we cleanup)
+            #and we don't really care too much about this
+            for l in (log, log, log, log, log.warn):
+                try:
+                    with xsync:
+                        cleanup_all_event_receivers()
+                        #all went well, we're done
+                        break
+                except Exception as e:
+                    l("failed to remove event receivers: %s", e)
         if self.fake_xinerama:
             cleanup_fakeXinerama()
         with xswallow:
             clean_keyboard_state()
+        GTKServerBase.do_cleanup(self)
+        log("close_gdk_display_source()")
+        close_gdk_display_source()
 
 
     def get_uuid(self):
@@ -357,7 +379,7 @@ class X11ServerCore(GTKServerBase):
         sinfo = info.setdefault("server", {})
         sinfo.update({
             "type"                  : "Python/gtk/x11",
-            "fakeXinerama"          : self.fake_xinerama and bool(self.libfakeXinerama_so),
+            "fakeXinerama"          : bool(self.libfakeXinerama_so),
             "libfakeXinerama"       : self.libfakeXinerama_so or "",
             })
         log("X11ServerBase.do_get_info took %ims", (monotonic_time()-start)*1000)
@@ -368,7 +390,7 @@ class X11ServerCore(GTKServerBase):
         info = GTKServerBase.get_ui_info(self, proto, wids, *args)
         #this is added here because the server keyboard config doesn't know about "keys_pressed"..
         if not self.readonly:
-            with xsync:
+            with xlog:
                 info.setdefault("keyboard", {}).update({
                     "state"             : {
                         "keys_pressed"   : tuple(self.keys_pressed.keys())
@@ -391,17 +413,14 @@ class X11ServerCore(GTKServerBase):
                 "XTest"                 : X11Keyboard.hasXTest(),
                 })
         #randr:
-        try:
-            with xsync:
-                sizes = RandR.get_xrr_screen_sizes()
-                if self.randr and len(sizes)>=0:
-                    sinfo["randr"] = {
-                        ""          : True,
-                        "options"   : tuple(reversed(sorted(sizes))),
-                        "exact"     : self.randr_exact_size,
-                        }
-        except XError:
-            log("failed to query randr screen sizes", exc_info=True)
+        with xlog:
+            sizes = RandR.get_xrr_screen_sizes()
+            if self.randr and len(sizes)>=0:
+                sinfo["randr"] = {
+                    ""          : True,
+                    "options"   : tuple(reversed(sorted(sizes))),
+                    "exact"     : self.randr_exact_size,
+                    }
         return info
 
 
@@ -453,7 +472,7 @@ class X11ServerCore(GTKServerBase):
             other_ui_clients = [s.uuid for s in self._server_sources.values() if s!=server_source and s.ui_client]
             translate_only = len(other_ui_clients)>0
             with xsync:
-                server_source.set_keymap(self.keyboard_config, self.keys_pressed, force, translate_only)
+                server_source.set_keymap(self.keyboard_config, self.keys_pressed, force, translate_only)    #pylint: disable=access-member-before-definition
                 self.keyboard_config = server_source.keyboard_config
         finally:
             # re-enable via idle_add to give all the pending
@@ -490,13 +509,8 @@ class X11ServerCore(GTKServerBase):
 
     def get_cursor_image(self):
         #must be called from the UI thread!
-        try:
-            with xsync:
-                return X11Keyboard.get_cursor_image()
-        except XError as e:
-            cursorlog.error("Error getting cursor data:")
-            cursorlog.error(" %s", e)
-            return None
+        with xlog:
+            return X11Keyboard.get_cursor_image()
 
     def get_cursor_data(self):
         #must be called from the UI thread!
@@ -600,7 +614,7 @@ class X11ServerCore(GTKServerBase):
                 closest[distance] = (w, h)
                 continue            #size is too small/big for client
             if new_size:
-                ew,eh = new_size
+                ew,eh = new_size    #pylint: disable=unpacking-non-sequence
                 if (ew*eh<w*h)==bigger:
                     continue        #we found a better (smaller/bigger) candidate already
             new_size = w,h
@@ -903,7 +917,7 @@ class X11ServerCore(GTKServerBase):
         with xsync:
             if self.do_process_mouse_common(proto, wid, pointer):
                 self._update_modifiers(proto, wid, modifiers)
-                self.pointer_device.wheel_motion(button, distance/1000.0)
+                self.pointer_device.wheel_motion(button, distance/1000.0)   #pylint: disable=no-member
 
     def get_pointer_device(self, deviceid):
         #mouselog("get_pointer_device(%i) input_devices_data=%s", deviceid, self.input_devices_data)
@@ -950,7 +964,7 @@ class X11ServerCore(GTKServerBase):
         pos = self.root_window.get_pointer()[-3:-1]
         uuid = None
         if proto:
-            ss = self._server_sources.get(proto)
+            ss = self.get_server_source(proto)
             if ss:
                 uuid = ss.uuid
         if pos!=pointer[:2] or self.input_devices=="xi":
@@ -962,7 +976,7 @@ class X11ServerCore(GTKServerBase):
     def _update_modifiers(self, proto, wid, modifiers):
         if self.readonly:
             return
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss:
             if self.ui_driver and self.ui_driver!=ss.uuid:
                 return
@@ -976,13 +990,14 @@ class X11ServerCore(GTKServerBase):
         if self._process_mouse_common(proto, wid, pointer, deviceid):
             self.button_action(pointer, button, pressed, deviceid)
 
-    def button_action(self, _pointer, button, pressed, deviceid=-1, *args):
+    def button_action(self, pointer, button, pressed, deviceid=-1, *args):
         device = self.get_pointer_device(deviceid)
         assert device, "pointer device %s not found" % deviceid
         try:
             with xsync:
                 device.click(button, pressed, *args)
         except XError:
+            log("button_action(%s, %s, %s, %s, %s)", pointer, button, pressed, deviceid, args, exc_info=True)
             log.error("Error: failed (un)press mouse button %s", button)
             if button>=4:
                 log.error(" (perhaps your Xvfb does not support mousewheels?)")
@@ -1023,7 +1038,8 @@ class X11ServerCore(GTKServerBase):
             tx = x-minx
             ty = y-miny
             screenshot.paste(window_image, (tx, ty))
-        buf = StringIOClass()
+        from io import BytesIO
+        buf = BytesIO()
         screenshot.save(buf, "png")
         data = buf.getvalue()
         buf.close()
